/////////////////////////////////////////////////////////////////////////////
// AI.c - Source file for computer players
//
// Hearts Card Game
// Copyright (C) 2001-2002 Leo C. Singleton IV
// http://www.leosingleton.com/projects/hearts/
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; version 2 of the
// License. 
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Revision History
//
// 9/28/01 LCS - removed debugging and dead code
// 6/2/02 LCS - added code to AISelectCard_LowerThanHighest() to prevent
//		stupid playing of the Queen of Spades; modified
//		AISelectCard_DumpBad() to play King and Ace of Spades even if the
//		Queen has not been played (this is safe, because this function is
//		only called if the player has no cards in the suit led); added
//		additional functionality to AISelectCard_RandomLowest() to prevent
//		the computer player from leading the Queen of Spades; added
//		functionality to not play the queen of spades in
//		AISelectCard_Highest()
// 6/3/02 LCS - fixed bug in AISelectCard_LowerThanHighest(), replaced card
//		with highestCard; added #ifndef to disable error double-checking
//		(Currently, over 60,000 copies of this code have been downloaded,
//		with no reports of ever generating one of these errors, so it is a
//		pretty safe assumption that they can never occur.)
// 6/10/02 LCS - changed #include "MainForm.h" to "GameLogic.h"
////////////////////////////////////////////////////////////////////////////


#include "AI.h"
#include "GameLogic.h"	// for AI helper functions
#include "CardDraw.h"


// AI subroutines
static UInt8 AISelectCard_HighestClub(UInt8 player);
static UInt8 AISelectCard_RandomLowest(UInt8 player, Boolean bLowest,
									   Boolean queenPlayed);
static UInt8 AISelectCard_LowerThanHighest(UInt8 player, Boolean last);
static UInt8 AISelectCard_Highest(UInt8 player);
static UInt8 AISelectCard_DumpBad(UInt8 player, Boolean queenPlayed);
static UInt8 AISelectCard_DefaultAI(UInt8 player);
static UInt8 HighestInSuit(UInt8 player, UInt8 suit);
static UInt8 CardIfAvailable(UInt8 player, UInt8 card);
static UInt8 Cheater(UInt8 player);


/////////////////////////////////////////////////////////////////////////////
// AISelectPass()
// AI routine for selecting 3 cards to pass and stores the card numbers
//   (0-12) in the location pDest
extern void AISelectPass(UInt8 player, UInt8 *pDest)
{
	// TODO: add AI for "shooting the moon"

	UInt8 play, i;
	UInt8 *pFinish = pDest + 3;
	UInt8 *pPos = pDest;

	pDest[0] = 255;
	pDest[1] = 255;
	pDest[2] = 255;

	// dump queen of spades
	play = CardIfAvailable(player, SUIT_SPADES * 13 + CARD_QUEEN);
	if (play != 255)
	{
		*pPos = play;
		pPos++;
	}

	// dump ace of spades
	play = CardIfAvailable(player, SUIT_SPADES * 13 + CARD_ACE);
	if (play != 255)
	{
		*pPos = play;
		pPos++;
	}

	// dump king of spades
	play = CardIfAvailable(player, SUIT_SPADES * 13 + CARD_KING);
	if (play != 255)
	{
		*pPos = play;
		pPos++;
		if (pPos == pFinish)
			return;
	}

	// dump highest cards
	i = 12;
	for (;;)
	{
		if (pDest[0] != i && pDest[1] != i)
		{
			*pPos = i;
			pPos++;
			if (pPos == pFinish)
				return;
		}
	
#ifndef NO_DOUBLE_CHECK
		if (i == 0)
		{
			// we should never get here - if so, it is a serious error
			ErrDisplay("AI Error: Unable to Proceed");
		}
#endif // NO_DOUBLE_CHECK

		i--;
	}
}


/////////////////////////////////////////////////////////////////////////////
// AISelectCard()
// AI routine for selecting card to play
extern UInt8 AISelectCard(UInt8 player)
{
	// TODO: add AI for "shooting the moon"
	
	Boolean bFirstTrick, bLeadPlayer, bLastPlayer, bCardsInSuit,
		bPointCardsInTrick, bQueenPlayed;
	UInt8 play = 0;
	UInt8 i, j;

	// set the variables that will be used to determine the AI to be used
	bFirstTrick = Game.Tricks == 0;
	bLeadPlayer = Game.Play[player - 1] == 255;
	bLastPlayer = Game.Play[(player + 1) % 4] != 255;
	bCardsInSuit = false;
	bQueenPlayed = false;
	bPointCardsInTrick = false;
	for (i=0; i<13; i++)
	{
		if (!Game.Used[player][i] && GetSuit(Game.Cards[player][i]) ==
			Game.SuitLed)
			bCardsInSuit = true;
		for (j=0; j<4; j++)
		{
			if (Game.Cards[j][i] == SUIT_SPADES * 13 + CARD_QUEEN)
				bQueenPlayed = Game.Used[j][i];
		}
	}
	for (i=0; i<4; i++)
	{
		if (Game.Play[i] != 255)
		{
			j = Game.Cards[i][Game.Play[i]]; // get value of any card played
			if (GetSuit(j) == SUIT_HEARTS || (GetSuit(j) == SUIT_SPADES &&
				GetCard(j) == CARD_QUEEN))
				bPointCardsInTrick = true;
		}
	}

	// call the appropriate AI routine
	if (bFirstTrick)
	{
		if (bLeadPlayer)
			play = 0;	// play the 2 of clubs (always the first card)
		else
		{
			if (bCardsInSuit)
				play = AISelectCard_HighestClub(player);
			else
				play = AISelectCard_RandomLowest(player, false,
					bQueenPlayed);
		}
	}
	else
	{
		play = Cheater(player);
		if (play == 255)
		{
			if (bLeadPlayer)
				play = AISelectCard_RandomLowest(player, true, bQueenPlayed);
			else
			{
				if (bCardsInSuit)
				{
					if (bLastPlayer)
					{
						if (bPointCardsInTrick)
							play = AISelectCard_LowerThanHighest(player, true);
						else
							play = AISelectCard_Highest(player);
					}
					else
						play = AISelectCard_LowerThanHighest(player, false);
				}
				else
					play = AISelectCard_DumpBad(player, bQueenPlayed);
			}
		}
	}

	// now, this is fairly redundant, but will prevent the computer from
	// accidentally cheating.  If play is invalid, we'll use the default AI,
	// which does not work well, but always yields a safe play
	if (!IsValidPlay(Game.Cards[player][play]) || Game.Used[player][play])
		play = AISelectCard_DefaultAI(player);

	return play;
}


/////////////////////////////////////////////////////////////////////////////
// AISelectCard_HighestClub()
// AI subroutine for selecting the highest club
// (subroutine of AISelectCard())
static UInt8 AISelectCard_HighestClub(UInt8 player)
{
	UInt8 play = 0;
	UInt8 i;

	for (i=0; i<13; i++)
	{
		if (GetSuit(Game.Cards[player][i]) == SUIT_CLUBS)
			play = i;
	}
	
	return play;
}


/////////////////////////////////////////////////////////////////////////////
// AISelectCard_RandomLowest()
// AI subroutine for selecting the lowest card in a randomly selected suit
// (subroutine of AISelectCard())
static UInt8 AISelectCard_RandomLowest(UInt8 player, Boolean bLowest,
									   Boolean queenPlayed)
{
	UInt8 cardFromSuit[4];
	UInt8 i;

	// clear card array
	for (i=0; i<4; i++)
		cardFromSuit[i] = 255;

	// fill card array
	if (bLowest)
	{
		// with lowest cards
		for (i=0; i<13; i++)
		{
			if (!Game.Used[player][i] &&
				cardFromSuit[GetSuit(Game.Cards[player][i])] == 255 &&
				Game.Cards[player][i] != SUIT_SPADES * 13 + CARD_QUEEN)
				cardFromSuit[GetSuit(Game.Cards[player][i])] = i;
		}
	}
	else
	{
		// with highest cards
		for (i=0; i<13; i++)
		{
			if (!Game.Used[player][i])
				cardFromSuit[GetSuit(Game.Cards[player][i])] = i;
		}
	}

	// let's eliminate stupid first moves
	if (cardFromSuit[SUIT_SPADES] < 13)
	{
		if (GetCard(Game.Cards[player][cardFromSuit[SUIT_SPADES]]) ==
			CARD_QUEEN)
			cardFromSuit[SUIT_SPADES] = 255;
	}

	if (cardFromSuit[SUIT_SPADES] < 13)
	{
		if (!queenPlayed && bLowest &&
			(GetCard(Game.Cards[player][cardFromSuit[SUIT_SPADES]]) == CARD_KING ||
			GetCard(Game.Cards[player][cardFromSuit[SUIT_SPADES]]) == CARD_ACE))
			cardFromSuit[SUIT_SPADES] = 255;
	}

	if (cardFromSuit[SUIT_HEARTS] < 13)
	{
		if (GetCard(Game.Cards[player][cardFromSuit[SUIT_HEARTS]]) == CARD_JACK ||
			GetCard(Game.Cards[player][cardFromSuit[SUIT_HEARTS]]) == CARD_QUEEN ||
			GetCard(Game.Cards[player][cardFromSuit[SUIT_HEARTS]]) == CARD_KING ||
			GetCard(Game.Cards[player][cardFromSuit[SUIT_HEARTS]]) == CARD_ACE)
			cardFromSuit[SUIT_HEARTS] = 255;
	}

	// TODO: make sure we don't lead a king or ace

	if (Game.HeartsBroken && Rand(4) == 2)
	{
		// 1 in 4 chance of getting here - play a heart
		if (cardFromSuit[SUIT_HEARTS] != 255)
			return cardFromSuit[SUIT_HEARTS];
	}

	if (Rand(3) == 1)
	{
		// 3/4 * 1/3 = 1 in 4 chance of getting here - play a spade
		if (cardFromSuit[SUIT_SPADES] != 255)
			return cardFromSuit[SUIT_SPADES];
	}

	if (Rand(2) == 1)
	{
		// 3/4 * 2/3 * 1/2 = 1 in 4 chance of getting here - play clubs
		if (cardFromSuit[SUIT_CLUBS] != 255)
			return cardFromSuit[SUIT_CLUBS];
	}

	// diamonds are left - play one - if no diamonds, then go into default AI
	if (cardFromSuit[SUIT_DIAMONDS] != 255)
		return cardFromSuit[SUIT_DIAMONDS];
	else
		return AISelectCard_DefaultAI(player);
}


/////////////////////////////////////////////////////////////////////////////
// AISelectCard_LowerThanHighest()
// AI subroutine for selecting the card lower than the highest played of the
//   suit led
// (subroutine of AISelectCard())
static UInt8 AISelectCard_LowerThanHighest(UInt8 player, Boolean last)
{
	UInt8 highestCard = 0;
	UInt8 i, card, play = 255;

	// find highest card
	for (i=0; i<4; i++)
	{
		if (Game.Play[i] != 255)
		{
			card = Game.Cards[i][Game.Play[i]];
			if (GetSuit(card) == Game.SuitLed && card > highestCard)
				highestCard = card;
		}
	}

	// find highest card lower than "card"
	// but don't play the queen of spades
	for (i=0; i<13; i++)
	{
		card = Game.Cards[player][i];
		if (!Game.Used[player][i] && GetSuit(card) == Game.SuitLed &&
			card < highestCard && card != SUIT_SPADES * 13 + CARD_QUEEN)
			play = i;
	}

	if (play == 255)
	{
		if (!last)
		{
			// if no card was found, try for lowest card
			for (i=0; i<13; i++)
			{
				if (!Game.Used[player][i] && GetSuit(Game.Cards[player][i])
					== Game.SuitLed)
					return i;
			}

			// still no card found - use default AI
			play = AISelectCard_DefaultAI(player);
		}
		else
			play = AISelectCard_Highest(player);
	}

	return play;
}


/////////////////////////////////////////////////////////////////////////////
// AISelectCard_Highest()
// AI subroutine for selecting the highest card in the suit led
// (subroutine of AISelectCard())
static UInt8 AISelectCard_Highest(UInt8 player)
{
	UInt8 i, play = 0;

	for (i=0; i<13; i++)
	{
		if (GetSuit(Game.Cards[player][i]) == Game.SuitLed &&
			!Game.Used[player][i] && Game.Cards[player][i] != SUIT_SPADES *
			13 + CARD_QUEEN)
			play = i;
	}
	
	return play;
}


/////////////////////////////////////////////////////////////////////////////
// AISelectCard_DumpBad()
// AI subroutine for selecting the worst card
// (subroutine of AISelectCard())
static UInt8 AISelectCard_DumpBad(UInt8 player, Boolean queenPlayed)
{
	UInt8 play;

	// dump queen of spades
	play = CardIfAvailable(player, SUIT_SPADES * 13 + CARD_QUEEN);
	if (play != 255)
		return play;

	// ace of spades
	play = CardIfAvailable(player, SUIT_SPADES * 13 + CARD_ACE);
	if (play != 255)
		return play;

	// king of spades
	play = CardIfAvailable(player, SUIT_SPADES * 13 + CARD_KING);
	if (play != 255)
		return play;

	// highest heart
	play = HighestInSuit(player, SUIT_HEARTS);
	if (play != 255)
		return play;

	// highest spade
	play = HighestInSuit(player, SUIT_SPADES);
	if (play != 255)
		return play;

	// highest diamond
	play = HighestInSuit(player, SUIT_DIAMONDS);
	if (play != 255)
		return play;

	// highest club
	play = HighestInSuit(player, SUIT_CLUBS);
	if (play != 255)
		return play;

	// we should never get here, but if we do, jump to default AI
	return AISelectCard_DefaultAI(player);
}


/////////////////////////////////////////////////////////////////////////////
// AISelectCard_DefaultAI()
// When all else fails, find any play
// (subroutine of AISelectCard())
static UInt8 AISelectCard_DefaultAI(UInt8 player)
{
	UInt8 i;
	
	// look for a card in the suit led
	if (Game.SuitLed != 255)
	{
		for (i=0; i<13; i++)
		{
			if (GetSuit(Game.Cards[player][i]) == Game.SuitLed &&
				!Game.Used[player][i])
				return i;
		}
	}

	for (i=0; i<13; i++)
	{
		if (!Game.Used[player][i])
			return i;
	}

#ifndef NO_DOUBLE_CHECK
	ErrDisplay("AI Error: Unable to proceed");
#endif // NO_DOUBLE_CHECK

	return 0;
}


/////////////////////////////////////////////////////////////////////////////
// HighestInSuit()
// Finds the highest card in the specified suit
static UInt8 HighestInSuit(UInt8 player, UInt8 suit)
{
	UInt8 play = 255;
	UInt8 i;

	for (i=0; i<13; i++)
	{
		if (GetSuit(Game.Cards[player][i]) == suit && !Game.Used[player][i])
			play = i;
	}
	
	return play;
}


/////////////////////////////////////////////////////////////////////////////
// CardIfAvailable()
// Returns the position of the specified card if it is held
static UInt8 CardIfAvailable(UInt8 player, UInt8 card)
{
	UInt8 i;

	for (i=0; i<13; i++)
	{
		if (Game.Cards[player][i] == card && !Game.Used[player][i])
			return i;
	}
	
	return 255;
}


/////////////////////////////////////////////////////////////////////////////
// Cheater()
// Tries to dump the queen on the human player.  Always.  Then hearts.
static UInt8 Cheater(UInt8 player)
{
	UInt8 i;
	UInt8 score = 0;

	// if the human has already played, then we can't cheat - we already know
	// which card he played
	if (Game.Play[0] != 255)
		return 255;

	// if suit was led and it's not spades, ignore
	if (Game.SuitLed == SUIT_SPADES || Game.SuitLed == 255)
	{
		UInt8 hasQueen = 255;
		UInt8 lowSpade = 255;

		// if we don't have the queen, then ignore
		for (i=0; i<13; i++)
		{
			if (!Game.Used[player][i] &&
				Game.Cards[player][i] == SUIT_SPADES * 13 + CARD_QUEEN)
				hasQueen = i;
		}
	
		if (hasQueen != 255)
		{
			// find the human player's lowest spade
			for (i=0; i<13; i++)
			{
				if (!Game.Used[0][i] && GetSuit(Game.Cards[0][i]) ==
					SUIT_SPADES && lowSpade == 255)
					lowSpade = Game.Cards[0][i];
			}

			// if spade is a king or ace, play the queen
			if (lowSpade != 255 && lowSpade > SUIT_SPADES * 13 + CARD_QUEEN)
				return hasQueen;
		}
	}


	// if the player appears to be shooting the moon, don't dump hearts
	for (i=1; i<3; i++)
		score += Game.RoundScore[i];
	if (score == 0 && ((Game.RoundScore[0] > 4 && Game.RoundScore[0] < 13) ||
		Game.RoundScore[0] > 16))
		return 255;

	
	// for more fun ;-), try to dump hearts on the player too
	if (Game.SuitLed == SUIT_HEARTS || (Game.SuitLed == 255 &&
		Game.HeartsBroken))
	{
		UInt8 lowHeart = 255;
		UInt8 lowerThanLowest = 255;
	
		// find human's lowest heart
		for (i=0; i<13; i++)
		{
			if (!Game.Used[0][i] && GetSuit(Game.Cards[0][i]) == SUIT_HEARTS
				&& lowHeart == 255)
				lowHeart = Game.Cards[0][i];
		}

		if (lowHeart != 255)
		{
			// find a heart lower than the player's lowest
			for (i=0; i<13; i++)
			{
				if (!Game.Used[player][i] &&
					GetSuit(Game.Cards[player][i]) == SUIT_HEARTS &&
					Game.Cards[player][i] < lowHeart)
					lowerThanLowest = i;
			}

			if (lowerThanLowest != 255)
				return lowerThanLowest;
		}
	}

	return 255;
}

