/////////////////////////////////////////////////////////////////////////////
// MainForm.c - Source file for application's main form
//
// Hearts Card Game
// Copyright (C) 2001-2002 Leo C. Singleton IV
// http://www.leosingleton.com/projects/hearts/
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; version 2 of the
// License. 
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Revision History
//
// 8/8/01 LCS - changed scrCopy to winPaint to upgrade from
//		Palm OS SDK 3.1 to SDK 4.0
// 8/10/01 LCS - added How to Play to menu.  Removed New Game from menu
// 9/5/01 LCS - replaced color searching routine with BACKGROUND_COLOR
//		constant.  Fixes background color problem with HandSpring Prisms.
// 9/28/01 LCS - removed debugging and dead code
// 12/30/01 LCS - added special handling for wide characters in player names
// 5/31/02 LCS - rewrote algorithm to print vertical names to fix bug in
//		wide character support
// 6/2/02 LCS - fixed bug in vertical display of certain "wide" (in pixels)
//		characters
// 6/3/02 LCS - wrote a new algorithm for DealCards(), which more randomly
//		distributes cards.  Although the difference is not noticeable, this
//		should reduce the occurrence of special cases that lead to bugs
//		(such as the all Hearts/Queen of Spades bug)
// 6/4/02 LCS - fixed bug that caused no playable card when the player was
//		dealt all Hearts/Queen of Spades
// 6/6/02 LCS - added code to support marking playable cards with dots
// 6/10/02 LCS - split MainForm.c into two files (GameLogic.c), to reduce
//		the size of MainForm.c (which was over 2100 lines); fixed warning
//		messages
/////////////////////////////////////////////////////////////////////////////


#include "AI.h"
#include "CardDraw.h"
#include "GameLogic.h"
#include "Hearts.h"
#include "MainForm.h"


// Local function declarations
static Boolean MainFormOpenEvent();
static Boolean MainFormCloseEvent();
static Boolean MainMenuEvent(UInt16 menuID);
static Boolean MainFormPenDownEvent(UInt16 x, UInt16 y);
static Boolean MainFormPenMoveEvent(UInt16 x, UInt16 y);
static Boolean MainFormPenUpEvent(UInt16 x, UInt16 y);
static Boolean MainFormCardSelect(UInt8 card);
static Boolean MainCtlSelectEvent(UInt16 controlID);
static Boolean MainFormAnimate();
static void DoPrefsDialog();
static Boolean MenuGetVisible();
static void UpdateTopLabel();
static Boolean AreYouSure();

// Game variables
static UInt8 PenDownOnCard = 255;
static Boolean PenOnCard = false;
static WinHandle AnimSaveWin = NULL, AnimBuffer = NULL;
static WinHandle MainScreenBuffer = NULL;
static UInt16 CardScrollPos = 0;
static Boolean IgnoreTaps = true;

// Game routines
static void DeallocateAnimationBuffers();
static void DrawPlayerCards(UInt8 start, UInt8 end, Boolean invert);
static UInt8 MapPointToCard(UInt16 x, UInt16 y);
static void MakeRectInPlay(RectangleType *pRect);
static void DrawPlayedCards();

static void MarkPlayableCards();
static void UnmarkPlayableCards();


/////////////////////////////////////////////////////////////////////////////
// MainFormHandleEvent()
// Event Handler for MainForm
extern Boolean MainFormHandleEvent(EventPtr event)
{
	Boolean handled = false;

	switch (event->eType)
	{
	case frmOpenEvent:
		handled = MainFormOpenEvent();
		break;
	case frmCloseEvent:
		handled = MainFormCloseEvent();
		break;
	case menuEvent:
		handled = MainMenuEvent(event->data.menu.itemID);
		break;
	case penDownEvent:
		handled = MainFormPenDownEvent(event->screenX, event->screenY);
		break;
	case penMoveEvent:
		handled = MainFormPenMoveEvent(event->screenX, event->screenY);
		break;
	case penUpEvent:
		handled = MainFormPenUpEvent(event->data.penUp.end.x,
			event->data.penUp.end.y);
		break;
	case ctlSelectEvent:
		handled = MainCtlSelectEvent(event->data.ctlSelect.controlID);
		break;
	case nilEvent:
		handled = MainFormAnimate();
		break;
	case frmUpdateEvent:
		PaintScreen(true);
		handled = false;	// make sure the title bar gets repainted too
		break;
	default:	// ignore all other events
		break;
	}

	return handled;
}


/////////////////////////////////////////////////////////////////////////////
// MainFormOpenEvent()
// Main form's initialization routine
// (subroutine of MainFormHandleEvent())
static Boolean MainFormOpenEvent()
{
	FormType *form;
	UInt16 index;

	// seed random number generator
	SysRandom(TimGetTicks());

	// draw main form
	form = FrmGetActiveForm();
	FrmHideObject(form, FrmGetObjectIndex(form, MainPassButton));
	FrmDrawForm(form);

	if (Game.GameStatus == GAMESTATUS_SELECTPASS)
	{
		// show/hide pass button
		index = FrmGetObjectIndex(form, MainPassButton);
		if (Game.CardPassCount == 3)
			FrmShowObject(form, index);
		else
			FrmHideObject(form, index);
	}
	
	PaintScreen(true);

	CheckLocks();

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// MainFormCloseEvent()
// Main form's destruction routine
// (subroutine of MainFormHandleEvent())
static Boolean MainFormCloseEvent()
{
	FormType *form;
	FieldType *status;

	// set status field handle to NULL to prevent it from destroying it
	form = FrmGetActiveForm();
	status = FrmGetObjectPtr(form, FrmGetObjectIndex(form, MainStatusLabel));
	FldSetTextHandle(status, NULL);

	DeallocateAnimationBuffers();
	
	return true;
}


/////////////////////////////////////////////////////////////////////////////
// MainMenuEvent()
// Event handler for menu events
// (subroutine of MainFormHandleEvent())
static Boolean MainMenuEvent(UInt16 menuID)
{
	Boolean handled = false;

	switch (menuID)
	{
	case MainGameScore:
		// View Score clicked
		DoScoreDialog();
		handled = true;
		break;

	case MainGameEndGame:
		// End Game
		if (AreYouSure())
			EndGame();
		handled = true;
		break;

	case MainOptionsPreferences:
		// Preferences clicked
		DoPrefsDialog();
		handled = true;
		break;

	case MainOptionsHowToPlay:
		// How to Play
		FrmHelp(HowToPlay);
		IgnoreTaps = true;
		handled = true;
		break;
	
	case MainOptionsAboutLeo:
		// About leosolutions.com
		FrmHelp(AboutLeoSolutions);
		IgnoreTaps = true;
		handled = true;
		break;
	
	case MainOptionsAbout:
		// About menu clicked
		FrmAlert(AboutAlert);
		IgnoreTaps = true;
		handled = true;
		break;

	default:
		break;
	}

	return handled;
}


/////////////////////////////////////////////////////////////////////////////
// MainFormPenDownEvent()
// Event handler for pen events
// (subroutine of MainFormHandleEvent())
static Boolean MainFormPenDownEvent(UInt16 x, UInt16 y)
{
	if (IgnoreTaps)
		return false;

	// sometimes menu taps can accidentally call this routine - ignore them
	if (MenuGetVisible())
		return false;
	
	// make sure we are at a point in the game for selecting cards
	if (Game.GameStatus != GAMESTATUS_SELECTPASS &&
		Game.GameStatus != GAMESTATUS_SELECTCARD)
		return false;
	
	PenDownOnCard = MapPointToCard(x, y);

	if (PenDownOnCard != 255)
	{
		// a card was clicked - highlight it
		DrawPlayerCards(PenDownOnCard, PenDownOnCard, true);
		PenOnCard = true;

		return true;
	}
	
	return false;
}


/////////////////////////////////////////////////////////////////////////////
// MainFormPenMoveEvent()
// Event handler for pen events
// (subroutine of MainFormHandleEvent())
static Boolean MainFormPenMoveEvent(UInt16 x, UInt16 y)
{
	UInt8 card;

	if (IgnoreTaps)
		return false;

	// if no card was selected, this event isn't for us
	if (PenDownOnCard == 255)
		return false;
	
	// determine if we are currently on a card
	card = MapPointToCard(x, y);

	// update inverting if needed
	if ((card == 255 || card != PenDownOnCard) && PenOnCard == true)
	{
		DrawPlayerCards(PenDownOnCard, PenDownOnCard, false);
		PenDownOnCard = 255;
		PenOnCard = false;
	}

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// MainFormPenUpEvent()
// Event handler for pen events
// (subroutine of MainFormHandleEvent())
static Boolean MainFormPenUpEvent(UInt16 x, UInt16 y)
{
	UInt8 card;

	if (IgnoreTaps)
		return false;

	// sometimes menu taps can accidentally call this routine - ignore them
	if (MenuGetVisible())
		return false;

	// handle new game taps
	if (Game.GameStatus == GAMESTATUS_NOGAME)
		NewGame();

	// if no card was selected, this message wasn't for us
	if (PenDownOnCard == 255)
		return false;

	// determine if we are currently on a card
	card = MapPointToCard(x, y);

	// update inverting if needed
	if ((card == 255 || card != PenDownOnCard) && PenOnCard == true)
	{
		DrawPlayerCards(PenDownOnCard, PenDownOnCard, false);
		PenDownOnCard = 255;
		PenOnCard = false;
	}

	if (card == PenDownOnCard)
	{
		// card was clicked

		// turn inversion off
		DrawPlayerCards(PenDownOnCard, PenDownOnCard, false);
		PenDownOnCard = 255;
		PenOnCard = false;

		MainFormCardSelect(card);
	}

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// MainFormCardSelect()
// Event handler for card selections
// (subroutine of MainFormPenUpEvent())
static Boolean MainFormCardSelect(UInt8 card)
{
#ifdef COLOR
	IndexedColorType /*newColor,*/ oldColor;
#endif // COLOR

	if (Game.GameStatus == GAMESTATUS_SELECTPASS)
	{
		FormType *form;
		UInt16 index;

		if (!Game.Invert[card] && Game.CardPassCount < 3)
		{
			// add card to 3
			Game.CardPassCount++;
		}
		else if (Game.Invert[card])
		{
			// remove card from 3
			Game.CardPassCount--;
		}
		else return false;

		// show/hide pass button
		form = FrmGetActiveForm();
		index = FrmGetObjectIndex(form, MainPassButton);
		if (Game.CardPassCount == 3)
			FrmShowObject(form, index);
		else
		{
#ifdef COLOR
			RectangleType rect;
#endif // COLOR
			FrmHideObject(form, index);
#ifdef COLOR
			oldColor = WinSetBackColor(BACKGROUND_COLOR);
			rect.topLeft.x = 40;
			rect.topLeft.y = 40;
			rect.extent.x = 80;
			rect.extent.y = 60;
			WinEraseRectangle(&rect, 0);
			WinSetBackColor(oldColor);
#endif // COLOR
		}

		// invert card
		Game.Invert[card] = !Game.Invert[card];
		DrawPlayerCards(card, card, false);
	}
	else
	{
		// GameStatus == GAMESTATUS_SELECTCARD
		if (IsValidPlay(Game.Cards[0][card]))
		{
			RectangleType rect;
			
			// set card as used
			Game.Used[0][card] = true;
			Game.Play[0] = card;
			
			// set suit led if this was the first play
			if (Game.SuitLed == 255)
				Game.SuitLed = GetSuit(Game.Cards[0][card]);

			// set HeartsBroken
			if (GetSuit(Game.Cards[0][card]) == SUIT_HEARTS)
				Game.HeartsBroken = true;

			// increment turn
			Game.Turn = 1;

			// clear card drawing area & redraw cards
			ShowMessage(MsgBlank);
			rect.topLeft.x = PLAYERCARDS_X + (card * 10);
			rect.topLeft.y = PLAYERCARDS_Y;
			rect.extent.x = 20;
			rect.extent.y = 30;
#ifdef COLOR
			oldColor = WinSetBackColor(BACKGROUND_COLOR);
#endif // COLOR
			WinEraseRectangle(&rect, 0);
#ifdef COLOR
			WinSetBackColor(oldColor);
#endif // COLOR

			if (card == 12)
				DrawPlayerCards(card - 1, card, false);
			else if (card == 0)
				DrawPlayerCards(card, card + 1, false);
			else
				DrawPlayerCards(card - 1, card + 1, false);

			// start animation
			Game.AnimStartX = PLAYERCARDS_X + (card * 10);
			Game.AnimStartY = PLAYERCARDS_Y;
			Game.AnimEndX = PLAYER0CARD_X;
			Game.AnimEndY = PLAYER0CARD_Y;
			Game.AnimLastX = Game.AnimStartX;
			Game.AnimLastY = Game.AnimStartY;
			Game.AnimProgress = 0;
			Game.AnimCard = GetCard(Game.Cards[0][card]);
			Game.AnimSuit = GetSuit(Game.Cards[0][card]);
			Game.GameStatus = GAMESTATUS_ANIMATE;
		}
	}

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// MainCtlSelectEvent()
// Event handler for button taps
// (subroutine of MainFormHandleEvent())
static Boolean MainCtlSelectEvent(UInt16 controlID)
{
	Boolean handled = false;
	UInt8 i, j;

	if (controlID == MainPassButton)
	{
		UInt8 PlayerA[3], PlayerB[3];
		FormType *form;
		
		// hide the pass button
		form = FrmGetActiveForm();
		FrmHideObject(form, FrmGetObjectIndex(form, MainPassButton));

		// load the human's cards into PlayerA
		i = 0;
		j = 0;
		while (i < 13 && j < 3)
		{
			if (Game.Invert[i])
			{
				PlayerA[j] = i;
				j++;
			}
			i++;
		}

		// TODO: This routine doesn't work!  But the player won't notice the
		//   difference - the cards were randomly drawn, so it doesn't matter
		//   who passes to whom
		switch (Game.PassTo)
		{
		case 0:
			// pass left
			AISelectPass(1, PlayerB);
			SwapCards(0, PlayerA, 1, PlayerB);
			break;

		case 1:
			// pass right
			AISelectPass(3, PlayerB);
			SwapCards(0, PlayerA, 3, PlayerB);
			break;

		case 2:
			// pass across
			AISelectPass(2, PlayerB);
			SwapCards(0, PlayerA, 2, PlayerB);
			break;
		}

		Game.PassTo++;

		StartPlay();

		handled = true;
	}

	return handled;
}


/////////////////////////////////////////////////////////////////////////////
// MainFormAnimate()
// Event handler for animation
// (subroutine of MainFormHandleEvent())
static Boolean MainFormAnimate()
{
	RectangleType rect;
	UInt16 err;
	WinHandle oldH;
	UInt16 x, y;

#ifdef COLOR
	IndexedColorType oldColor;
#endif // COLOR

	// this is a simple workaround to ignore that first tap that occurs when
	// the application is opened
	IgnoreTaps = false;
	
	// pause animation if a menu is visible
	if (MenuGetVisible())
		return false;

	// pause animation if another form is on top
	if (FrmGetFormId(FrmGetActiveForm()) != MainForm)
		return false;

	if (Game.GameStatus == GAMESTATUS_ANIMATE)
	{
		// update animation progress
		Game.AnimProgress += Prefs.AnimationSpeed;

		// if the background buffers have not been created, create them
		if (AnimSaveWin == NULL || AnimBuffer == NULL)
		{
			rect.topLeft.x = 0;
			rect.topLeft.y = 0;
			rect.extent.x = 160;
			rect.extent.y = 160;
			AnimSaveWin = (WinHandle)WinCreateOffscreenWindow(160, 160,
				screenFormat, &err);
			if (err)
				return false;
			WinCopyRectangle(0, AnimSaveWin, &rect, 0, 0, winPaint);

			AnimBuffer = (WinHandle)WinCreateOffscreenWindow(160, 160,
				screenFormat, &err);
			if (err)
				return false;
			// copying the background to the animation buffer should prevent
			// the white pixels that appear around animated cards on color
			// displays
			WinCopyRectangle(0, AnimBuffer, &rect, 0, 0, winPaint);
		}

		// calculate new X and Y positions
		if (Game.AnimProgress < 200)
		{
			x = ((Game.AnimEndX - Game.AnimStartX) * Game.AnimProgress / 200)
				+ Game.AnimStartX;
			y = ((Game.AnimEndY - Game.AnimStartY) * Game.AnimProgress / 200)
				+ Game.AnimStartY;
		}
		else
		{
			x = Game.AnimEndX;
			y = Game.AnimEndY;
		}

		// copy background to buffer at old location
		rect.topLeft.x = Game.AnimLastX;
		rect.topLeft.y = Game.AnimLastY;
		rect.extent.x = 20;
		rect.extent.y = 30;
		MakeRectInPlay(&rect);

		WinCopyRectangle(AnimSaveWin, AnimBuffer, &rect, rect.topLeft.x,
			rect.topLeft.y, winPaint);

		// draw card to buffer at new location
		oldH = WinSetDrawWindow(AnimBuffer);
		DrawCard(Game.AnimCard, Game.AnimSuit, CARDTYPE_FULL, false, x, y);
		WinSetDrawWindow(oldH);

		// copy buffer to current window
		WinCopyRectangle(AnimBuffer, 0, &rect, rect.topLeft.x,
			rect.topLeft.y,	winPaint);

		rect.topLeft.x = x;
		rect.topLeft.y = y;
		rect.extent.x = 20;
		rect.extent.y = 30;
		MakeRectInPlay(&rect);
		WinCopyRectangle(AnimBuffer, 0, &rect, rect.topLeft.x,
			rect.topLeft.y,	winPaint);

		// current x & y are now the old ones
		Game.AnimLastX = x;
		Game.AnimLastY = y;

		if (Game.AnimProgress >= 200)
		{
			// animation is complete

			// destroy buffers
			DeallocateAnimationBuffers();

			// do whatever happens next
			PostAnimate();
		}
	}
	else if (Game.GameStatus == GAMESTATUS_NOGAME)
	{
		// this is the opening screen with scrolling cards
		UInt16 cardpos;
		UInt8 i;

		if (MainScreenBuffer == NULL)
		{
			// create background buffer 
			MemHandle BitmapH;
			BitmapType *pBitmap;

			rect.topLeft.x = 0;
			rect.topLeft.y = 0;
			rect.extent.x = 160;
			rect.extent.y = 160;
			MainScreenBuffer = (WinHandle)WinCreateOffscreenWindow(160, 160,
				screenFormat, &err);
			if (err)
				return true;

			// draw the background bitmap
			BitmapH = (MemHandle)DmGetResource(bitmapRsc, MainScreenBitmap);
			if (BitmapH)
			{
				pBitmap = (BitmapType*)MemHandleLock(BitmapH);
				oldH = WinSetDrawWindow(MainScreenBuffer);
				WinDrawBitmap(pBitmap, 0, 16);
				MemHandleUnlock(BitmapH);
				WinSetDrawWindow(oldH);
			}
		}

		// make the cards dance
		CardScrollPos = (CardScrollPos + (Prefs.AnimationDelay / 2)) % 458;

		oldH = WinSetDrawWindow(MainScreenBuffer);
		
#ifdef COLOR
		oldColor = WinSetBackColor(BACKGROUND_COLOR);
#endif // COLOR

		// erase old cards
		rect.topLeft.x = 0;
		rect.topLeft.y = 16;
		rect.extent.x = 160;
		rect.extent.y = 30;
		WinEraseRectangle(&rect, 0);
		rect.topLeft.y = 105;
		WinEraseRectangle(&rect, 0);
		rect.topLeft.y = 46;
		rect.extent.x = 20;
		rect.extent.y = 59;
		WinEraseRectangle(&rect, 0);
		rect.topLeft.x = 140;
		WinEraseRectangle(&rect, 0);

#ifdef COLOR
		WinSetBackColor(oldColor);
#endif // COLOR

		// draw new cards
		for (i=0; i<13; i++)
		{
			cardpos = (CardScrollPos + (i * 35)) % 458;
			if (cardpos < 140)
			{
				// moving right along the top
				x = cardpos;
				y = 16;
			}
			else if (cardpos < 229)
			{
				// moving down along the right
				x = 140;
				y = cardpos - 124;
			}
			else if (cardpos < 369)
			{
				// moving left along the bottom
				x = 369 - cardpos;
				y = 105;
			}
			else
			{
				// moving up along the left
				x = 0;
				y = 474 - cardpos;
			}
			
			DrawCard(i, SUIT_HEARTS, CARDTYPE_FULL, false, x, y);
		}
		WinSetDrawWindow(oldH);

		// copy buffer to screen
		rect.topLeft.x = 0;
		rect.topLeft.y = 16;
		rect.extent.x = 160;
		rect.extent.y = 144;
		WinCopyRectangle(MainScreenBuffer, 0, &rect, rect.topLeft.x,
			rect.topLeft.y,	winPaint);
	}

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// DoScoreDialog()
// Shows score dialog box
extern void DoScoreDialog()
{
	FormType *form;
	FieldType *name[4], *score[4], *total[4];
	MemHandle nameH[4], scoreH[4], totalH[4];
	char *pStr;
	UInt8 i;

	// init dialog
	form = FrmInitForm(ScoreDialog);

	// get pointers to fields
	name[0] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreName1Field));
	name[1] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreName2Field));
	name[2] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreName3Field));
	name[3] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreName4Field));
	score[0] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreScore1Field));
	score[1] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreScore2Field));
	score[2] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreScore3Field));
	score[3] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreScore4Field));
	total[0] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreTotal1Field));
	total[1] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreTotal2Field));
	total[2] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreTotal3Field));
	total[3] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		ScoreTotal4Field));

	for (i=0; i<4; i++)
	{
		// alloc mem for name
		nameH[i] = MemHandleNew(StrLen(Prefs.Names[i]) + 1);
		pStr = MemHandleLock(nameH[i]);
		StrCopy(pStr, Prefs.Names[i]);
		MemHandleUnlock(nameH[i]);
		
		// allocate memory for round scores
		scoreH[i] = MemHandleNew(8);
		pStr = MemHandleLock(scoreH[i]);
		StrIToA(pStr, Game.RoundScore[i]);
		MemHandleUnlock(scoreH[i]);

		// allocate memory for total scores
		totalH[i] = MemHandleNew(8);
		pStr = MemHandleLock(totalH[i]);
		StrIToA(pStr, Game.TotalScore[i]);
		MemHandleUnlock(totalH[i]);

		// set text handles
		FldSetTextHandle(name[i], nameH[i]);
		FldSetTextHandle(score[i], scoreH[i]);
		FldSetTextHandle(total[i], totalH[i]);
	}

	// show dialog
	FrmDoDialog(form);

	// destroy form - this will also take care of field handles
	FrmDeleteForm(form);

	// ignore the next tap (caused by the continue button) - we don't want
	// to accidentally select any cards
	IgnoreTaps = true;
}


/////////////////////////////////////////////////////////////////////////////
// DoPrefsDialog()
// Shows preferences dialog box
static void DoPrefsDialog()
{
	FormType *form;
	MemHandle nameH[4];
	FieldType *field[4];
	ListType *listSpeed, *listSmooth;
	ControlType *trigSpeed, *trigSmooth;
	UInt8 i;
	char *pStr1, *pStr2;
	UInt8 speedPos = 1, smoothPos = 1;

	// init dialog
	form = FrmInitForm(PrefsDialog);

	// get pointers to fields
	field[0] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsYourNameField));
	field[1] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsOpponent1NameField));
	field[2] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsOpponent2NameField));
	field[3] = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsOpponent3NameField));

	// allocate memory for names and copy from the existing names
	for (i=0; i<4; i++)
	{
		nameH[i] = MemHandleNew(26);
		pStr2 = MemHandleLock(nameH[i]);
		StrCopy(pStr2, Prefs.Names[i]);
		MemHandleUnlock(nameH[i]);
		FldSetTextHandle(field[i], nameH[i]);
	}

	// get old values
	switch (Prefs.AnimationDelay)
	{
	case PREF_ANIMSMOOTH_LOW:
		smoothPos = 0;
		break;
	case PREF_ANIMSMOOTH_MED:
		smoothPos = 1;
		break;
	case PREF_ANIMSMOOTH_HIGH:
		smoothPos = 2;
	}
	switch (Prefs.AnimationSpeed / Prefs.AnimationDelay)
	{
	case PREF_ANIMSPEED_SLOW:
		speedPos = 0;
		break;
	case PREF_ANIMSPEED_MED:
		speedPos = 1;
		break;
	case PREF_ANIMSPEED_FAST:
		speedPos = 2;
	}

	// set animation speed drop-down box
	listSpeed = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsAnimSpeedList));
	trigSpeed = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsAnimSpeedTrigger));
	LstSetSelection(listSpeed, speedPos);
	CtlSetLabel(trigSpeed, LstGetSelectionText(listSpeed, speedPos));

	// set animation smoothness drop-down box
	listSmooth = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsAnimSmoothList));
	trigSmooth = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
		PrefsAnimSmoothTrigger));
	LstSetSelection(listSmooth, smoothPos);
	CtlSetLabel(trigSmooth, LstGetSelectionText(listSmooth, smoothPos));

	// show dialog
	if (FrmDoDialog(form) == PrefsOKButton)
	{
		// OK was selected

		for (i=0; i<4; i++)
		{
			// trash old name handles and use new ones
			pStr1 = MemHandleLock(nameH[i]);
			StrCopy(Prefs.Names[i], pStr1);
			MemHandleUnlock(nameH[i]);
		}

		switch (LstGetSelection(listSmooth))
		{
		case 0:
			Prefs.AnimationDelay = PREF_ANIMSMOOTH_LOW;
			break;
		case 1:
			Prefs.AnimationDelay = PREF_ANIMSMOOTH_MED;
			break;
		case 2:
			Prefs.AnimationDelay = PREF_ANIMSMOOTH_HIGH;
			break;
		}
		
		switch (LstGetSelection(listSpeed))
		{
		case 0:
			Prefs.AnimationSpeed = Prefs.AnimationDelay *
				PREF_ANIMSPEED_SLOW;
			break;
		case 1:
			Prefs.AnimationSpeed = Prefs.AnimationDelay *
				PREF_ANIMSPEED_MED;
			break;
		case 2:
			Prefs.AnimationSpeed = Prefs.AnimationDelay *
				PREF_ANIMSPEED_FAST;
			break;
		}

		PaintScreen(true);
	}

	// destroy form
	FrmDeleteForm(form);

	// ignore the next tap (caused by the continue button) - we don't want
	// to accidentally select any cards
	IgnoreTaps = true;
}


/////////////////////////////////////////////////////////////////////////////
// MenuGetVisible()
// Returns true if a menu is currently visible
static Boolean MenuGetVisible()
{
	MenuBarType *menuBarTypeP = MenuGetActiveMenu();
	Boolean menuVisible = false;
	
	if (menuBarTypeP)
		menuVisible = menuBarTypeP->attr.visible;

	return menuVisible;
}


/////////////////////////////////////////////////////////////////////////////
// UpdateTopLabel()
// Updates top label with status information
static void UpdateTopLabel()
{
	FormType *form;
	FieldType *status;
	DmResID id;
	MemHandle strH;
	
	form = FrmGetActiveForm();
	status = FrmGetObjectPtr(form, FrmGetObjectIndex(form, MainStatusLabel));

	switch (Game.GameStatus)
	{
	case GAMESTATUS_NOGAME:
		id = StatusTapAnywhere;
		break;

	default:
		id = StatusLeoSolutions;
		break;
	}
	
	strH = DmGetResource(strRsc, id);
	if (!strH)
		return;

	FldSetTextHandle(status, strH);
	FldDrawField(status);
}


/////////////////////////////////////////////////////////////////////////////
// AreYouSure()
// "Are you sure you want to end the current game?" dialog box
static Boolean AreYouSure()
{
	if (FrmAlert(EndGameAlert) == 0)
		return true;
	else
		return false;
}


/////////////////////////////////////////////////////////////////////////////
// DeallocateAnimationBuffers()
// Releases memory used by drawing buffers
static void DeallocateAnimationBuffers()
{
	if (AnimBuffer != NULL)
		WinDeleteWindow(AnimBuffer, false);
	if (AnimSaveWin != NULL)
		WinDeleteWindow(AnimSaveWin, false);
	AnimBuffer = NULL;
	AnimSaveWin = NULL;
}


/////////////////////////////////////////////////////////////////////////////
// DeallocateMainScreenBuffers()
// Releases memory used by main screen drawing buffers
extern void DeallocateMainScreenBuffers()
{
	if (MainScreenBuffer != NULL)
		WinDeleteWindow(MainScreenBuffer, false);
	MainScreenBuffer = NULL;
}


/////////////////////////////////////////////////////////////////////////////
// DrawPlayerCards()
// Draws the cards held by the player
extern void DrawPlayerCardsAll()
{
	DrawPlayerCards(0, 12, false);
}

static void DrawPlayerCards(UInt8 start, UInt8 end, Boolean invert)
{
	UInt8 type, i;

	// draw cards
	for (i=start; i<end+1; i++)
	{
		if (!Game.Used[0][i] && i != Game.CurrentSelection)
		{
			if (i == 12)
				type = CARDTYPE_FULL;
			else
			{
				if (!Game.Used[0][i + 1])
					type = CARDTYPE_LEFT;
				else
					type = CARDTYPE_FULL;
			}
			DrawCard(Game.Cards[0][i] % 13, Game.Cards[0][i] / 13, type,
				invert ^ Game.Invert[i], (i * 10) + PLAYERCARDS_X,
				PLAYERCARDS_Y);
		}
	}
}


/////////////////////////////////////////////////////////////////////////////
// ShowMessage()
// Loads the specified string ID and displays it at the bottom of the screen
extern void ShowMessage(UInt16 stringID)
{
	MemHandle strH;
	char *pStr;
	UInt16 length, width;
	RectangleType rect;

#ifdef COLOR
	IndexedColorType newColor, oldBackColor, oldForeColor;
	RGBColorType newRGB;
#endif // COLOR

	// get a pointer to the string data
	strH = DmGetResource(strRsc, stringID);
	if (!strH)
		return;
	pStr = (char*)MemHandleLock(strH);

	// calculate width of string
	length = StrLen(pStr);
	width = FntLineWidth(pStr, length);
	
	// erase the old string
#ifdef COLOR
	newRGB.r = 255;
	newRGB.g = 255;
	newRGB.b = 255;
	newColor = WinRGBToIndex(&newRGB);
	oldForeColor = WinSetTextColor(newColor);
	oldBackColor = WinSetBackColor(BACKGROUND_COLOR);
#endif // COLOR

	rect.topLeft.x = 0;
	rect.topLeft.y = MESSAGE_Y;
	rect.extent.x = 160;
	rect.extent.y = 160-MESSAGE_Y;
	WinEraseRectangle(&rect, 0);
	
	// draw the string
	WinDrawChars(pStr, length, 80 - (width / 2), MESSAGE_Y);

#ifdef COLOR
	WinSetTextColor(oldForeColor);
	WinSetBackColor(oldBackColor);
#endif // COLOR

	// release pointer
	MemHandleUnlock(strH);
	DmReleaseResource(strH);
}


/////////////////////////////////////////////////////////////////////////////
// MapPointToCard()
// Returns the card at the specified x-y coordinates
static UInt8 MapPointToCard(UInt16 x, UInt16 y)
{
	UInt8 card;
	
	// see if click was inside card borders
	if (x < PLAYERCARDS_X || y < PLAYERCARDS_Y || x > PLAYERCARDS_X + 140 ||
		y > PLAYERCARDS_Y + 30)
		return 255;

	// determine which card was pressed
	card = (UInt8)((x - PLAYERCARDS_X) / 10);
	if (card > 12)
		card = 12;
	
	// the card that was clicked is gone
	if (Game.Used[0][card])
	{
		// maybe it's the card underneath
		if (card == 0)
			return 255;
		card--;
		if (Game.Used[0][card])
			return 255;		// nope - an empty space was clicked
	}

	return card;
}


/////////////////////////////////////////////////////////////////////////////
// MakeRectInPlay()
// Makes sure a rectangle is within the playing field before doing a blit
// (This prevents system elements from being overwritten and allows for the
//  effect of cards flying off the screen)
static void MakeRectInPlay(RectangleType *pRect)
{
	if (pRect->topLeft.x < FIELD_LEFT)
	{
		pRect->extent.x -= FIELD_LEFT - pRect->topLeft.x;
		pRect->topLeft.x = FIELD_LEFT;
	}

	if (pRect->topLeft.y < FIELD_TOP)
	{
		pRect->extent.y -= FIELD_TOP - pRect->topLeft.y;
		pRect->topLeft.y = FIELD_TOP;
	}

	if (pRect->topLeft.x + pRect->extent.x - 1 > FIELD_RIGHT)
	{
		pRect->extent.x = FIELD_RIGHT - pRect->topLeft.x + 1;
	}

	if (pRect->topLeft.y + pRect->extent.y - 1 > FIELD_BOTTOM)
	{
		pRect->extent.y = FIELD_BOTTOM - pRect->topLeft.y + 1;
	}
}


/////////////////////////////////////////////////////////////////////////////
// DrawPlayerNames()
// Draws the names of the four players on the screen
extern void DrawPlayerNames()
{
	RectangleType rect;
	UInt16 length, width, height;
	FontID oldFont;
	UInt8 i, c;
	char *pStr;
	UInt8 x, y;

#ifdef COLOR
	IndexedColorType newColor, oldBackColor, oldForeColor;
	RGBColorType newRGB;
#endif // COLOR

#ifdef COLOR
	newRGB.r = 255;
	newRGB.g = 255;
	newRGB.b = 255;
	newColor = WinRGBToIndex(&newRGB);
	oldForeColor = WinSetTextColor(newColor);
	oldBackColor = WinSetBackColor(BACKGROUND_COLOR);
#endif // COLOR

	// select the big font
	oldFont = FntSetFont(largeFont);
	
	rect.topLeft.x = 0;
	rect.topLeft.y = 0;
	rect.extent.x = PLAYERNAME_MAXHEIGHT;
	rect.extent.y = PLAYERNAME_MAXWIDTH;

	// loop through each player
	for (i=0; i<4; i++)
	{
		pStr = Prefs.Names[i];
		length = StrLen(pStr);
		width = FntLineWidth(pStr, length);
	
		if (i == 0 || i == 2)
		{
			// write these horizontally
			WinDrawChars(pStr, length, 80 - (width / 2),
				i == 0 ? PLAYER0NAME_Y : PLAYER2NAME_Y);
		}
		else
		{
			// new routine for wide character support - added 5/31/02

			// note the difference between the offset and chars variables
			UInt16 offset = 0;	// number of bytes past start of buffer
			UInt16 charlength = 0;	// this is similar to length, but in
									// characters for wide character support
			WChar wch;			// temporary variable for wide characters
			UInt16 charSize;	// size of a wide character (either 1 or 2
								// bytes)

			// count the number of characters
			if (WideCharacterSupport)
			{
				while (offset < length)
				{
					offset += TxtGetNextChar(pStr, offset, &wch);
					charlength++;
				}
			}
			else
			{
				// this Palm OS device doesn't support wide characters, so
				// the number of bytes must be the same as the number of
				// characters
				charlength = length;
			}

			// truncate length if necessary
			if (charlength > PLAYERNAME_MAXLENGTHALL)
				charlength = PLAYERNAME_MAXLENGTHALL;

			// the next few lines haven't changed
			height = FntCharHeight();
			if (i == 1)
				y = PLAYER1NAME_YC - (height * charlength / 2);
			else
				y = PLAYER3NAME_YC - (height * charlength / 2);

			// print the vertical names character by character
			offset = 0;
			charSize = 0;
			for (c=0; c<charlength; c++)
			{
				if (WideCharacterSupport)
				{
					// get the next wide character's width
					offset += charSize;
					charSize = TxtGetNextChar(pStr, offset, &wch);

					// MAX_CHAR_WIDTH fixes a bug with some characters not
					// being printed in wide character formats (due to the
					// characters being so wide they fall past the edge of
					// the screen).  This way, we just cram all the
					// characters on there anyway.
					width = FntCharWidth(pStr[offset]);
					if (width > MAX_CHAR_WIDTH)
						width = MAX_CHAR_WIDTH;
				}
				else
				{
					// each byte is a char; this is easy
					offset = c;
					charSize = 1;
					
					// see comment above on MAX_CHAR_WIDTH
					width = FntCharWidth(pStr[c]);
					if (width > MAX_CHAR_WIDTH)
						width = MAX_CHAR_WIDTH;
				}

				if (i == 1)
					x = PLAYER1NAME_XC - (width / 2);
				else
					x = PLAYER3NAME_XC - (width / 2);

				// Note: Don't use WinDrawChar here!!!  It's not compatible
				// with some older ROMs.
				WinDrawChars(pStr + offset, charSize, x, y);

				y += height;
			}
		}
	}

	// restore the original font
	FntSetFont(oldFont);

#ifdef COLOR
	WinSetTextColor(oldForeColor);
	WinSetBackColor(oldBackColor);
#endif // COLOR
}


/////////////////////////////////////////////////////////////////////////////
// PaintScreen()
// Repaints the entire screen
extern void PaintScreen(Boolean erase)
{
#ifdef COLOR
	IndexedColorType oldColor;
#endif // COLOR

	if (erase)
	{
		RectangleType rect;
		FormType *form;
	
		// clear the old screen
#ifdef COLOR
		oldColor = WinSetBackColor(BACKGROUND_COLOR);
#endif // COLOR
		rect.topLeft.x = FIELD_LEFT;
		rect.topLeft.y = FIELD_TOP;
		rect.extent.x = FIELD_RIGHT - FIELD_LEFT + 1;
		rect.extent.y = FIELD_BOTTOM - FIELD_TOP + 1;
		WinEraseRectangle(&rect, 0);
#ifdef COLOR
		WinSetBackColor(oldColor);
#endif // COLOR

		// we may have erased the pass button, so repaint it
		form = FrmGetActiveForm();
		CtlDrawControl((ControlType*)FrmGetObjectPtr(form,
			FrmGetObjectIndex(form, MainPassButton)));
	}
	
	UpdateTopLabel();

	switch (Game.GameStatus)
	{
	case GAMESTATUS_NOGAME:
		// display splash screen
		MainFormAnimate();
		IgnoreTaps = true;
		break;

	case GAMESTATUS_SELECTPASS:
	case GAMESTATUS_SELECTCARD:
	case GAMESTATUS_ANIMATE:
		// display cards
		DrawPlayerCardsAll();
		DrawPlayerNames();
		if (Game.GameStatus == GAMESTATUS_SELECTCARD)
			ShowMessage(MsgSelectCardPlay);
		else
		{
			switch (Game.PassTo)
			{
			case 0:
				ShowMessage(MsgPassLeft);
				break;
			case 1:
				ShowMessage(MsgPassRight);
				break;
			case 2:
				ShowMessage(MsgPassAcross);
				break;
			}
		}

		DrawPlayedCards();

		break;
	}

	// if enabled in the preferences, mark playable cards with dots
	if (Prefs.MarkPlayableCards)
	{
		if (Game.GameStatus == GAMESTATUS_SELECTCARD)
			MarkPlayableCards();
		else
			UnmarkPlayableCards();
	}
}


/////////////////////////////////////////////////////////////////////////////
// DrawPlayedCards()
// Draw any cards that have been played but not yet erased
static void DrawPlayedCards()
{
	UInt8 card, i, x = 0, y = 0;

	// this simple check creates awkward displays for a second, but prevents
	// fragments left behind on the screen
	if (Game.Play[Game.Turn] != 255)
		return;

	for (i=0; i<4; i++)
	{
		if (Game.Play[i] != 255)
		{
			card = Game.Cards[i][Game.Play[i]];
			switch (i)
			{
			case 0:
				x = PLAYER0CARD_X;
				y = PLAYER0CARD_Y;
				break;
			case 1:
				x = PLAYER1CARD_X;
				y = PLAYER1CARD_Y;
				break;
			case 2:
				x = PLAYER2CARD_X;
				y = PLAYER2CARD_Y;
				break;
			case 3:
				x = PLAYER3CARD_X;
				y = PLAYER3CARD_Y;
				break;
			}
			DrawCard(GetCard(card), GetSuit(card), CARDTYPE_FULL, false, x,
				y);
		}
	}
}


/////////////////////////////////////////////////////////////////////////////
// MarkPlayableCards()
// Places a dot over playable cards
static void MarkPlayableCards()
{
	Boolean playable[4];
	UInt8 card;

	// determine playable suits
	// since we've already got a fancy algorithm to compute this, simply
	// reuse it with test cases
	playable[SUIT_CLUBS] = IsValidPlay(SUIT_CLUBS * 13 + CARD_2);
	playable[SUIT_DIAMONDS] = IsValidPlay(SUIT_DIAMONDS * 13 + CARD_2);
	playable[SUIT_SPADES] = IsValidPlay(SUIT_SPADES * 13 + CARD_2);
	playable[SUIT_HEARTS] = IsValidPlay(SUIT_HEARTS * 13 + CARD_2);

	// place dots over playable cards
	for (card=0; card<13; card++)
	{
	}
}


/////////////////////////////////////////////////////////////////////////////
// UnmarkPlayableCards()
// Erases the dot over playable cards
static void UnmarkPlayableCards()
{
	// draw solid box over dot area
}

